#
# OpenVMP, 2023
#
# Author: Roman Kuzmenko
# Created: 2023-12-30
#
# Licensed under Apache License, Version 2.0.

import importlib
import os
import shutil
import subprocess
import json

from . import runtime_python
from . import logging as pc_logging


class CondaPythonRuntime(runtime_python.PythonRuntime):
    def __init__(self, ctx, version=None):
        super().__init__(ctx, "conda", version)

        self.conda_path = shutil.which("conda")
        if self.conda_path is None:
            self.conda_cli = importlib.import_module("conda.cli.python_api")
            self.conda_cli.run_command("config", "--quiet")
            info_json, _, _ = self.conda_cli.run_command("info", "--json")
            info = json.loads(info_json)
            if "CONDA_EXE" in info["env_vars"]:
                self.conda_path = info["env_vars"]["CONDA_EXE"]
            else:
                root_prefix = info["root_prefix"]
                root_bin = os.path.join(root_prefix, "bin")
                root_scripts = os.path.join(root_prefix, "Scripts")
                search_paths = [
                    root_scripts,
                    root_bin,
                    root_prefix,
                ]
                if os.name == "nt":
                    search_path_strings = ";".join(search_paths)
                else:
                    search_path_strings = ":".join(search_paths)
                self.conda_path = shutil.which(
                    "conda",
                    path=search_path_strings,
                )

    def once(self):
        with self.lock:
            if not self.initialized:
                with pc_logging.Action("Conda", "create", self.version):
                    if self.conda_path is None:
                        raise Exception(
                            "ERROR: PartCAD is configured to use conda, but conda is missing"
                        )

                    try:
                        os.makedirs(self.path)

                        # Install new conda environment with the preferred Python version
                        p = subprocess.Popen(
                            [
                                self.conda_path,
                                "create",
                                "-y",
                                "-q",
                                "--json",
                                "-p",
                                self.path,
                                "python=%s" % self.version,
                            ],
                            stdout=subprocess.PIPE,
                            stderr=subprocess.PIPE,
                        )
                        _, stderr = p.communicate()
                        if not stderr is None and stderr != b"":
                            pc_logging.error(
                                "conda env install error: %s" % stderr
                            )

                        # Install pip into the newly created conda environment
                        p = subprocess.Popen(
                            [
                                self.conda_path,
                                "install",
                                "-y",
                                "-q",
                                "--json",
                                "-p",
                                self.path,
                                "pip",
                            ],
                            stdout=subprocess.PIPE,
                            stderr=subprocess.PIPE,
                        )
                        _, stderr = p.communicate()
                        if not stderr is None and stderr != b"":
                            pc_logging.error(
                                "conda pip install error: %s" % stderr
                            )

                        self.initialized = True
                    except Exception as e:
                        shutil.rmtree(self.path)
                        raise e

    async def run(self, cmd, stdin="", cwd=None):
        self.once()

        return await super().run(
            [
                self.conda_path,
                "run",
                "--no-capture-output",
                "-p",
                self.path,
                "python" if os.name != "nt" else "pythonw",
                # "python%s" % self.version,  # This doesn't work on Windows
            ]
            + cmd,
            stdin,
            cwd=cwd,
        )



#
# OpenVMP, 2024
#
# Author: Roman Kuzmenko
# Created: 2024-04-20
#
# Licensed under Apache License, Version 2.0.
#

from .interface_inherit import InterfaceInherits
from .interface import Interface
from . import logging as pc_logging


class WithPorts(Interface):
    interfaces: dict[str, dict[str, dict[str, str]]]

    def __init__(
        self,
        name: str,
        project,
        config: dict = {},
    ):
        super().__init__(name, project, config, config_section="implements")

        self.interfaces = None

    def get_interfaces(self):
        if self.interfaces is None:
            self.instantiate_interfaces()
        return self.interfaces

    def get_interface(self, interface_name: str):
        return self.get_interfaces()[interface_name]

    def instantiate_interfaces(self):
        self.interfaces = {}

        # Recursively merge the inherited interfaces
        def merge_inherits(
            inherits, interface_state: str = "", top_level=False
        ):
            if (
                not top_level
                and len(inherits.keys()) == 1
                and (len(list(inherits.values())[0].instances.keys()) == 1)
            ):
                compatible = True
            else:
                compatible = False

            for interface_name, inherit in inherits.items():
                interface = inherit.interface

                # Ignore abstract interfaces
                if interface.abstract:
                    continue

                if not ":" in interface_name:
                    interface_name = self.project.name + ":" + interface_name

                if not compatible and interface_name not in self.interfaces:
                    self.interfaces[interface_name] = {}

                for instance_name in inherit.instances.keys():
                    if instance_name != "" and interface_state != "":
                        instance_full_name = (
                            interface_state + "-" + instance_name
                        )
                    elif instance_name != "":
                        instance_full_name = instance_name
                    elif interface_state != "":
                        instance_full_name = interface_state
                    else:
                        instance_full_name = ""

                    if not compatible:
                        if instance_name not in self.interfaces[interface_name]:
                            self.interfaces[interface_name][
                                instance_full_name
                            ] = {}

                        for port_name in interface.get_ports().keys():
                            if instance_full_name != "" and port_name != "":
                                port_full_name = (
                                    instance_full_name + "-" + port_name
                                )
                            elif instance_full_name != "":
                                port_full_name = instance_full_name
                            elif port_name != "":
                                port_full_name = port_name
                            else:
                                port_full_name = ""

                            self.interfaces[interface_name][instance_full_name][
                                port_name
                            ] = port_full_name

                    merge_inherits(interface.get_parents(), instance_full_name)

        merge_inherits(self.get_parents(), top_level=True)

    def info(self):
        return {
            "interfaces": dict(
                (
                    interface_name,
                    dict(
                        (
                            instance_name,
                            dict(
                                (port_name, port)
                                for port_name, port in instance.items()
                            ),
                        )
                        for instance_name, instance in interface.items()
                    ),
                )
                for interface_name, interface in self.get_interfaces().items()
            ),
            "ports": dict(
                (
                    port_name,
                    {
                        "location": port.location,
                        "sketch": f"{port.sketch.project_name}:{port.sketch.name}",
                    },
                )
                for port_name, port in self.get_ports().items()
            ),
        }



#
# OpenVMP, 2023
#
# Author: Roman Kuzmenko
# Created: 2023-12-30
#
# Licensed under Apache License, Version 2.0.

import os
import shutil

from . import runtime_python


class NonePythonRuntime(runtime_python.PythonRuntime):
    exec_name: str

    def __init__(self, ctx, version=None):
        super().__init__(ctx, "none", version)

        if os.name == "nt" and shutil.which("pythonw") is not None:
            self.exec_name = "pythonw"
        elif shutil.which("python3") is not None:
            self.exec_name = "python3"
        else:
            self.exec_name = "python"

        if not self.initialized:
            os.makedirs(self.path)
            self.initialized = True

    async def run(self, cmd, stdin="", cwd=None):
        return await super().run(
            [self.exec_name] + cmd,
            stdin,
            cwd=cwd,
        )



#
# OpenVMP, 2023
#
# Author: Roman Kuzmenko
# Created: 2023-08-19
#
# Licensed under Apache License, Version 2.0.
#

from . import project as p


class ImportConfiguration:
    def __init__(self, config_obj={}):
        self.config_obj = config_obj
        self.name = config_obj.get("name")
        if not "type" in config_obj:
            if "url" in config_obj:
                if config_obj["url"].endswith(".tar.gz"):
                    config_obj["type"] = "tar"
                else:
                    config_obj["type"] = "git"
            elif "path" in config_obj:
                config_obj["type"] = "local"
            else:
                raise ValueError("Import configuration type is not set")
        self.import_config_type = config_obj.get("type")
        self.import_config_is_root = config_obj.get("isRoot", False)
        self.include_paths = self.config_obj.get("includePaths", [])


class ProjectFactory(ImportConfiguration):
    def __init__(self, ctx, parent, import_config_obj):
        super().__init__(import_config_obj)
        self.ctx = ctx
        self.parent = parent

        if parent is None:
            self.config_path = ctx.config_path
            self.config_dir = ctx.config_dir
        else:
            self.config_path = parent.config_path
            self.config_dir = parent.config_dir

        # TODO(clairbee): Initialize the config object if necessary

    def _create(self, config):
        # TODO(clairbee): Finalize the config object if necessary
        self.project = p.Project(
            self.ctx, self.name, self.path, include_paths=self.include_paths
        )
        # Make the project config inherit some properties of the import config
        self.project.config_obj["type"] = self.import_config_type
        self.project.config_obj["isRoot"] = self.import_config_is_root

    def _save(self):
        self.ctx.projects[self.name] = self.project



#
# OpenVMP, 2024
#
# Author: Roman Kuzmenko
# Created: 2024-03-23
#
# Licensed under Apache License, Version 2.0.
#

import asyncio
from path import Path
import tempfile

from .ai import Ai
from . import logging as pc_logging
from .shape import Shape
from .user_config import user_config


class ShapeWithAi(Shape, Ai):

    def __init__(self, config):
        super().__init__(config)

    # @override
    async def get_summary_async(self, project=None):
        configured_summary = await super().get_summary_async(project)
        if not configured_summary is None:
            return configured_summary

        image_filename = tempfile.mktemp(".png")
        await self.render_png_async(project.ctx, project, image_filename)

        prompt = """The attached image is a single-color (the color doesn't
matter) line drawing of a mechanical design.
The design is stored in the folder "%s" and is named "%s".
""" % (
            self.project_name,
            self.name,
        )
        if not self.desc is None:
            prompt += (
                'The design is accompanied by the following description: "%s". '
                % self.desc
            )
        prompt += """Create a text which describes the design displayed on the
image so that a blind person (with background in mechanical engineering and
computer aided design) can picture it in their mind.
Do not repeat any information from the prompt without significant changes.
Make no assumptions. Provide as many details as possible.
Refer to the design as "this design" and not as "the image".
Produce text which is ready to be narrated as is.
"""

        config = {
            "model": (
                "gpt-4o"
                if not user_config.openai_api_key is None
                else "gemini-1.5-pro"
            ),
        }
        summary = self.generate(
            "Desc",
            self.project_name,
            self.name,
            prompt,
            config,
            image_filenames=[image_filename],
        )
        return summary[0] if len(summary) > 0 else "Failed to summarize"

    # @override
    def get_summary(self, project=None):
        return asyncio.run(self.get_summary_async(project))



#
# OpenVMP, 2024
#
# Author: Roman Kuzmenko
# Created: 2024-09-07
#
# Licensed under Apache License, Version 2.0.
#

import tempfile

from .provider_request_caps import ProviderRequestCaps
from .provider_request_order import ProviderRequestOrder
from .provider_request_quote import ProviderRequestQuote
from .provider_factory_python import ProviderFactoryPython
from .provider_data_cart import *
from . import logging as pc_logging

# from . import sync_threads as pc_thread


class ProviderFactoryManufacturer(ProviderFactoryPython):
    def __init__(self, ctx, source_project, target_project, config):
        with pc_logging.Action(
            "InitManuf", target_project.name, config["name"]
        ):
            super().__init__(
                ctx,
                source_project,
                target_project,
                config,
            )
            # Complement the config object here if necessary

            self._create(config)
            self.provider.is_part_available = self.is_part_available
            self.provider.load = self.load
            self.provider.query_caps = self.query_caps
            self.provider.query_quote = self.query_quote
            self.provider.query_order = self.query_order

    async def is_part_available(self, cart_item: ProviderCartItem):
        # TODO(clairbee): add vendor/SKU-based availability check
        caps = await self.provider.get_caps()
        if cart_item.material:
            if not cart_item.material in caps["materials"]:
                return False
            if cart_item.color:
                # TODO(clairbee): implement color mapping as a function in
                #                 the corresponding module
                #                 (resolve cart_item.color as a color object)
                if cart_item.color not in list(
                    map(
                        lambda c: c["name"],
                        caps["materials"][cart_item.material]["colors"],
                    )
                ):
                    return False
            if cart_item.finish:
                # TODO(clairbee): implement finish mapping as a function in
                #                 the corresponding module
                #                 (resolve cart_item.finish as a finish object)
                if cart_item.finish not in list(
                    map(
                        lambda f: f["name"],
                        caps["materials"][cart_item.material]["finishes"],
                    )
                ):
                    pc_logging.debug(
                        "Provider %s does not support finish %s"
                        % (self.name, cart_item.finish)
                    )
                    pc_logging.debug(
                        "Supported finishes: %s"
                        % caps["materials"][cart_item.material]["finishes"]
                    )
                    return False
        return True

    async def load(self, cart_item: ProviderCartItem):
        """Load the CAD binary into the cart item based on the provider capabilities."""
        caps = await self.provider.get_caps()
        # TODO(clairbee): Make the below more generic
        if "formats" in caps and "step" in caps["formats"]:
            part = self.ctx.get_part(cart_item.name)
            filepath = tempfile.mktemp(".step")
            await part.render_step_async(self.ctx, filepath=filepath)
            step = open(filepath, "rb").read()
            cart_item.add_binary("step", step)
        else:
            # TODO(clairbee): add support for other formats
            pc_logging.error(
                f"Provider {self.provider.name} does not support STEP format."
            )

    async def query_caps(self, request: ProviderRequestCaps):
        # TODO(clairbee): does it make sense to run this in a separate thread?
        # return await pc_thread.run_async(
        #   self.query_script, self.provider, "caps", request.compose()
        # )
        return await self.query_script(self.provider, "caps", request.compose())

    async def query_quote(self, request: ProviderRequestQuote):
        # TODO(clairbee): does it make sense to run this in a separate thread?
        # return await pc_thread.run_async(
        #   self.query_script, self.provider, "quote", request.compose(),
        # )
        return await self.query_script(
            self.provider, "quote", request.compose()
        )

    async def query_order(self, request: ProviderRequestOrder):
        # TODO(clairbee): does it make sense to run this in a separate thread?
        # return await pc_thread.run_async(
        #   self.query_script, self.provider, "order", request.compose(),
        # )
        return await self.query_script(
            self.provider, "order", request.compose()
        )



#
# OpenVMP, 2024
#
# Author: Roman Kuzmenko
# Created: 2024-04-20
#
# Licensed under Apache License, Version 2.0.
#

import cadquery as cq

from . import logging as pc_logging
from .sketch_factory_file import SketchFactoryFile


class SketchFactoryDxf(SketchFactoryFile):
    tolerance = 0.000001
    include = []
    exclude = []

    def __init__(self, ctx, source_project, target_project, config):
        with pc_logging.Action("InitDXF", target_project.name, config["name"]):
            super().__init__(
                ctx,
                source_project,
                target_project,
                config,
                extension=".dxf",
            )

            if "tolerance" in config:
                self.tolerance = float(config["tolerance"])

            if "include" in config:
                if isinstance(config["include"], list):
                    self.include = config["include"]
                elif isinstance(config["include"], str):
                    self.include = [config["include"]]

            if "exclude" in config:
                if isinstance(config["exclude"], list):
                    self.exclude = config["exclude"]
                elif isinstance(config["exclude"], str):
                    self.exclude = [config["exclude"]]

            self._create(config)

    async def instantiate(self, sketch):
        await super().instantiate(sketch)

        with pc_logging.Action("DXF", sketch.project_name, sketch.name):
            try:
                workplane = cq.importers.importDXF(
                    self.path,
                    tol=self.tolerance,
                    include=self.include,
                    exclude=self.exclude,
                )
                shape = workplane.val().wrapped
            except Exception as e:
                pc_logging.exception(
                    "Failed to import the DXF file: %s: %s" % (self.path, e)
                )
                shape = None

            self.ctx.stats_sketches_instantiated += 1

            return shape



#
# OpenVMP, 2024
#
# Author: Roman Kuzmenko
# Created: 2024-09-07
#
# Licensed under Apache License, Version 2.0.
#

import base64
import os
import pickle
import sys

sys.path.append(os.path.join(os.path.dirname(__file__), "wrappers"))
from cq_serialize import register as register_cq_helper

from .provider_factory_file import ProviderFactoryFile
from .runtime_python import PythonRuntime

from . import wrapper
from . import logging as pc_logging


class ProviderFactoryPython(ProviderFactoryFile):
    runtime: PythonRuntime
    cwd: str

    def __init__(
        self,
        ctx,
        source_project,
        target_project,
        config,
        can_create=False,
        python_version=None,
    ):
        super().__init__(
            ctx,
            source_project,
            target_project,
            config,
            extension=".py",
            can_create=can_create,
        )
        self.cwd = config.get("cwd", None)

        if python_version is None:
            # TODO(clairbee): stick to a default constant or configured version
            python_version = self.project.python_version
        if python_version is None:
            # Stay one step ahead of the minimum required Python version
            python_version = "3.10"
        if python_version == "3.12" or python_version == "3.11":
            pc_logging.debug(
                "Downgrading Python version to 3.10 to minimize compatibility issues"
            )
            python_version = "3.10"

        self.runtime = self.ctx.get_python_runtime(python_version)

    def info(self, provider):
        info: dict[str, object] = provider.shape_info()
        info.update(
            {
                "runtime_version": self.runtime.version,
                "runtime_path": self.runtime.path,
            }
        )
        return info

    async def prepare_script(self, provider) -> bool:
        """
        Finish initialization of PythonRuntime
        which was too expensive to do in the constructor
        """

        # Install dependencies of this package
        await self.runtime.prepare_for_package(self.project)
        await self.runtime.prepare_for_shape(self.config)

        return await super().prepare_script(provider)

    async def query_script(self, provider, script_name, request):
        extra = ""
        if script_name == "avail":
            vendor = request.get("vendor", None)
            sku = request.get("sku", None)
            if not vendor and not sku:
                extra = request["name"]
            else:
                if not vendor:
                    vendor = "None"
                if not sku:
                    sku = "None"
                extra = vendor + ":" + sku
        with pc_logging.Action(
            script_name.capitalize(),
            provider.project_name,
            provider.name,
            extra,
        ):
            prepared = await self.prepare_script(provider)
            if not prepared:
                pc_logging.error(
                    "Failed to prepare %s of %s" % (script_name, provider.name)
                )
                return None

            # Get the path to the wrapper script
            # which needs to be executed
            wrapper_path = wrapper.get("provider.py")

            # Build the request
            request["partcad_version"] = sys.modules["partcad"].__version__
            request["verbose"] = pc_logging.getLevel() <= pc_logging.DEBUG
            request["api"] = script_name
            request["parameters"] = {}
            if "parameters" in self.config:
                for param_name, param in self.config["parameters"].items():
                    # Check if this parameter has a value set
                    if "default" in param:
                        request["parameters"][param_name] = param["default"]

            # TODO(clairbee): Add support for patching. Copy files or drop runpy
            # patch = {}
            # if "patch" in self.config:
            #     patch.update(self.config["patch"])
            # request["patch"] = patch

            # Serialize the request
            register_cq_helper()
            picklestring = pickle.dumps(request)
            request_serialized = base64.b64encode(picklestring).decode()

            await self.runtime.ensure("ocp-tessellate")
            await self.runtime.ensure("numpy==1.24.1")
            await self.runtime.ensure("numpy-quaternion==2023.0.4")
            await self.runtime.ensure("nptyping==1.24.1")
            await self.runtime.ensure("cadquery")
            cwd = self.project.config_dir
            if self.cwd is not None:
                cwd = os.path.join(self.project.config_dir, self.cwd)
            response_serialized, errors = await self.runtime.run(
                [
                    wrapper_path,
                    os.path.abspath(self.path),
                    os.path.abspath(cwd),
                ],
                request_serialized,
            )
            if len(errors) > 0:
                error_lines = errors.split("\n")
                for error_line in error_lines:
                    provider.error("%s: %s" % (provider.name, error_line))

            try:
                response = base64.b64decode(response_serialized)
                register_cq_helper()
                result = pickle.loads(response)
            except Exception as e:
                provider.error(
                    "Exception while deserializing %s: %s" % (provider.name, e)
                )
                return None

            if "exception" in result:
                provider.error("%s: %s" % (provider.name, result["exception"]))
                return None

            self.ctx.stats_provider_queries += 1

            return result



#
# OpenVMP, 2024
#
# Author: Roman Kuzmenko
# Created: 2024-03-16
#
# Licensed under Apache License, Version 2.0.
#

# This script is executed within a python runtime environment
# (no need for a sandbox) to speed up parallel rendering

import os
import sys

import cadquery as cq

# import build123d as b3d

sys.path.append(os.path.dirname(__file__))
import wrapper_common


def process(path, request):
    try:
        cq_obj = cq.Solid.makeBox(1, 1, 1)
        cq_obj.wrapped = request["wrapped"]

        vertices, triangles = cq_obj.tessellate(
            request["tolerance"], request["angularTolerance"]
        )

        # b3d_obj = b3d.Shape(request["wrapped"])
        # vertices, triangles = b3d.Mesher._mesh_shape(
        #     b3d_obj, request["tolerance"], request["angularTolerance"]
        # )

        with open(path, "w") as f:
            f.write("# OBJ file\n")
            for v in vertices:
                f.write("v %.4f %.4f %.4f\n" % (v.x, v.y, v.z))
                # f.write("v %.4f %.4f %.4f\n" % (v[0], v[1], v[2]))
            for p in triangles:
                f.write("f")
                for i in p:
                    f.write(" %d" % (i + 1))
                f.write("\n")

        return {
            "success": True,
            "exception": None,
        }
    except Exception as e:
        wrapper_common.handle_exception(e)
        return {
            "success": False,
            "exception": e,
        }


path, request = wrapper_common.handle_input()

# Perform rendering
response = process(path, request)

wrapper_common.handle_output(response)



#
# OpenVMP, 2024
#
# Author: Roman Kuzmenko
# Created: 2024-03-23
#
# Licensed under Apache License, Version 2.0.
#

from .part_factory_cadquery import PartFactoryCadquery
from .part_factory_feature_ai import PartFactoryFeatureAi
from . import logging as pc_logging


class PartFactoryAiCadquery(PartFactoryCadquery, PartFactoryFeatureAi):
    def __init__(self, ctx, source_project, target_project, config):
        # Override the path determined by the parent class to enable "enrich"
        config["path"] = config["name"] + ".py"
        self.lang = self.LANG_PYTHON

        with pc_logging.Action("InitAiCq", target_project.name, config["name"]):
            PartFactoryFeatureAi.__init__(
                self,
                config,
                "cadquery",
                "CadQuery 2.0 script",
                """Generate a complete functioning script, not just a code snippet.
Import all the required modules (including 'math' and 'cadquery' itself).
Do not use methods: tetrahedron, hexahedron.
Do not generate comments.
Do not export anything.
Use "show_object()" to display the part.
""",
            )
            PartFactoryCadquery.__init__(
                self,
                ctx,
                source_project,
                target_project,
                config,
                can_create=True,
            )

            self.on_init_ai()



#
# OpenVMP, 2024
#
# Author: Roman Kuzmenko
# Created: 2024-04-20
#
# Licensed under Apache License, Version 2.0.
#

import asyncio
import math
import re

from .geom import Location
from .interface_inherit import InterfaceInherits
from .sketch import Sketch
from . import logging as pc_logging
from .utils import resolve_resource_path

from OCP.gp import (
    gp_Trsf,
    gp_Ax1,
    gp_Pnt,
    gp_Dir,
    gp_Vec,
)


class InterfacePort:
    """One of the ports provided by the interface,
    either explicitly (inside "ports:")
    or implicitly (inherited from "inherits:")."""

    name: str
    location: Location = None
    source_project_name: str
    source_sketch_name: str
    source_sketch_spec: str
    sketch: Sketch = None

    def __init__(
        self,
        name,
        project,
        config: dict = {},
        sketch: Sketch = None,
        location: Location = None,
    ):
        self.name = name

        if location is not None:
            self.location = location
        elif config.get("location", None) is not None:
            self.location = Location(config["location"])

        if sketch is not None:
            self.sketch = sketch
        elif "sketch" in config:
            if "project" in config:
                self.source_project_name = config["project"]
                if (
                    self.source_project_name == "this"
                    or self.source_project_name == ""
                ):
                    self.source_project_name = project.name
            else:
                self.source_project_name = project.name

            self.source_sketch_name = config["sketch"]
            if ":" in self.source_sketch_name:
                self.source_project_name, self.source_sketch_name = (
                    resolve_resource_path(
                        self.source_project_name,
                        self.source_sketch_name,
                    )
                )
                self.source_sketch_spec = (
                    self.source_project_name + ":" + self.source_sketch_name
                )
                self.sketch = project.ctx.get_sketch(self.source_sketch_spec)
            else:
                self.source_project_name = project.name
                self.source_sketch_spec = (
                    self.source_project_name + ":" + self.source_sketch_name
                )
                self.sketch = project.get_sketch(self.source_sketch_name)

    def __repr__(self):
        return f"<Port: {self.name}, location:{str(self.location)}>"


PARAM_MOVE = "move"
PARAM_TURN = "turn"


class InterfaceParameter:
    """One of the parameters provided by the interface,
    either explicitly (inside "parameters:")
    or implicitly (inherited from "inherits:")."""

    name: str
    dir: list[float]
    type: int = PARAM_MOVE
    min: float
    max: float
    default: float

    def __init__(self, config: dict = {}):
        self.name = config.get("name", "param")
        self.type = config.get("type", PARAM_MOVE)
        self.dir = config.get("dir", [1.0, 0.0, 0.0])
        self.min = config.get("min", 0.0)
        self.max = config.get("max", 0.0)
        self.default = config.get("default", 0.0)

    def __repr__(self):
        return f"<Parameter: {self.name}, default: {self.default}, min:{self.min}, max:{self.max}, dir:{self.dir}, type:{self.type}>"

    @staticmethod
    def config_normalize(config):
        if isinstance(config, (int, float)):
            config = {
                "min": config,
                "max": config,
                "default": config,
            }
        elif isinstance(config, list):
            new_config = {}
            new_config["min"] = config[0]
            if len(config) > 1:
                new_config["max"] = config[1]
                if len(config) > 2:
                    new_config["default"] = config[2]
            config = new_config

        elif not isinstance(config, dict):
            raise Exception("Invalid parameter configuration")

        if "default" in config:
            if "min" not in config:
                config["min"] = config["default"]
            if "max" not in config:
                config["max"] = config["default"]
        else:
            if "min" not in config:
                config["min"] = 0.0
            if "max" not in config:
                config["max"] = 0.0

            if config["min"] * config["max"] <= 0:
                config["default"] = 0.0
            else:
                config["default"] = (config["min"] + config["max"]) / 2.0

        return config

    @staticmethod
    def config_finalize(config):
        name = config.get("name", None)

        if name == "moveX":
            config["type"] = PARAM_MOVE
            config["dir"] = [1.0, 0.0, 0.0]
        elif name == "moveY":
            config["type"] = PARAM_MOVE
            config["dir"] = [0.0, 1.0, 0.0]
        elif name == "moveZ":
            config["type"] = PARAM_MOVE
            config["dir"] = [0.0, 0.0, 1.0]
        elif name == "turnX":
            config["type"] = PARAM_TURN
            config["dir"] = [1.0, 0.0, 0.0]
        elif name == "turnY":
            config["type"] = PARAM_TURN
            config["dir"] = [0.0, 1.0, 0.0]
        elif name == "turnZ":
            config["type"] = PARAM_TURN
            config["dir"] = [0.0, 0.0, 1.0]

        if not "type" in config:
            config["type"] = PARAM_MOVE

        return config

    def get_offsets(self, value):
        if self.min is not None and value < self.min:
            pc_logging.warning(
                "Parameter %s: value below minimum: %f" % (self.name, value)
            )
        if self.max is not None and value > self.max:
            pc_logging.warning(
                "Parameter %s: value above maximum: %f" % (self.name, value)
            )

        trsf = gp_Trsf()
        if self.type == PARAM_MOVE:
            if value != 0:
                trsf.SetTranslationPart(
                    gp_Vec(
                        self.dir[0] * value,
                        self.dir[1] * value,
                        self.dir[2] * value,
                    )
                )
                return [trsf]
        elif self.type == PARAM_TURN:
            if value != 0:
                trsf.SetRotation(
                    gp_Ax1(
                        gp_Pnt(),
                        gp_Dir(
                            self.dir[0],
                            self.dir[1],
                            self.dir[2],
                        ),
                    ),
                    value * math.pi / 180.0,
                )
                return [trsf]
        return []


# TODO(clairbee): introduce "Entity" as a shared parent to Shape and Interface
#                 to share "show()"
class Interface:
    """Stored as a singleton in the package and defines the interface.
    Explicitly contains all inherited ports and instances of sub-interfaces.
    """

    config: str
    config_section: str
    name: str
    full_name: str  # including project name
    desc: str
    abstract: bool
    lead_port: int

    ports: dict[str, InterfacePort]  # both own and inherited
    inherits: dict[str, InterfaceInherits] | None  # not set until instantiate()
    compatible_with: list[
        str
    ]  # list of ancestor interfaces with the same ports

    params: dict[str, InterfaceParameter]

    count: int

    def __init__(
        self,
        name: str,
        project,
        config: dict = {},
        config_section: str = "inherits",
    ):
        # TODO(clairbee): remove this circular dependency
        self.project = project

        self.config = config
        self.config_section = config_section
        self.name = name
        self.full_name = project.name + ":" + name
        self.desc = config.get("desc", "")
        self.desc = self.desc.strip() if self.desc is not None else ""
        self.abstract = config.get("abstract", False)
        self.lead_port = config.get("leadPort", None)

        self.ports = None
        self.inherits = None
        self.compatible_with = set()

        self.count = 0

        # pc_logging.debug("Initializing interface: %s" % name)

        # Initialize parameters space and freedom of movement
        # Not to be confused with specific parameter values.
        # See InterfaceInherits for values specific to a particular instance.
        self.params = {}
        params_config = config.get("parameters", None)
        if params_config is not None:
            if isinstance(params_config, list):
                params_config = {param: {} for param in params_config}
            elif not isinstance(params_config, dict):
                raise Exception(
                    "Invalid 'parameters' section in the interface '%s'"
                    % self.name
                )

            for param_name, param_config in params_config.items():
                param_config = InterfaceParameter.config_normalize(param_config)
                param_config["name"] = param_name
                param_config = InterfaceParameter.config_finalize(param_config)
                self.params[param_name] = InterfaceParameter(param_config)

        self.project.ctx.stats_interfaces += 1

    def get_ports(self):
        if self.ports is None:
            self.instantiate_ports()  # Fill in own ports
            self.instantiate()  # Get ports from parents
        return self.ports

    def instantiate_ports(self):
        self.ports = {}

        if self.config.get("ports", None) is not None:
            ports_config = self.config["ports"]
            if isinstance(ports_config, list):
                ports_config = {port_name: {} for port_name in ports_config}
            elif isinstance(ports_config, str):
                ports_config = {ports_config: {}}
            elif not isinstance(ports_config, dict):
                raise Exception(
                    "Invalid 'ports' section in the interface '%s'" % self.name
                )

            for port_name, port_config in ports_config.items():
                if isinstance(port_config, list):
                    port_config = {"location": port_config}
                self.ports[port_name] = InterfacePort(
                    port_name, self.project, port_config
                )

    def get_parents(self):
        if self.inherits is None:
            self.instantiate()
        return self.inherits

    def instantiate(self):
        self.project.ctx.stats_interfaces_instantiated += 1
        self.inherits = {}
        self.get_ports()  # Make sure self.ports is initialized

        # Initialize inheritance ("inherits" or "implements")
        inherits_config = self.config.get(self.config_section, None)
        if inherits_config is not None:
            if isinstance(inherits_config, str):
                inherits_config = {inherits_config: ""}  # {}???

            if len(inherits_config.keys()) == 1 and (
                isinstance(list(inherits_config.values())[0], str)
                or len(list(inherits_config.values())[0]) == 1
            ):
                compatible_with_parents = True
            else:
                compatible_with_parents = False

            for interface_name, interface_config in inherits_config.items():
                # Resolve the parameter values in the interface name
                def interface_template_resolve(m):
                    tmpl = m.group(1)
                    param_name = tmpl[0 : tmpl.index(":")]
                    value = self.params[param_name].default
                    if ":" in tmpl:
                        expr = tmpl[tmpl.index(":") + 1 :]
                        globals = {"__builtins__": {}}
                        locals = {"value": value}
                        value = eval(expr, globals, locals)

                    return value

                interface_name = re.sub(
                    "%([^%*]+)%",
                    interface_template_resolve,
                    interface_name,
                )

                inherit = InterfaceInherits(
                    interface_name, self.project, interface_config
                )
                if inherit.interface is None:
                    pc_logging.error(
                        "Failed to inherit interface: %s" % interface_name
                    )
                    continue
                self.inherits[inherit.name] = inherit

                if compatible_with_parents:
                    self.compatible_with.add(inherit.name)
                    self.compatible_with = self.compatible_with.union(
                        inherit.interface.compatible_with
                    )

                for (
                    instance_name,
                    instance_location,
                ) in inherit.instances.items():
                    # pc_logging.debug(
                    #     "Inherited ports: %s"
                    #     % str(inherit.interface.get_ports())
                    # )
                    for (
                        port_name,
                        port,
                    ) in inherit.interface.get_ports().items():
                        if instance_name != "":
                            inherited_port_name = (
                                instance_name + "-" + port_name
                            )
                        else:
                            inherited_port_name = port_name

                        if port.location is None:
                            port_location = instance_location
                        else:
                            trsf = port.location.wrapped.Transformation()
                            # pc_logging.debug(
                            #     "Instance location: %s" % instance_location
                            # )
                            trsf.PreMultiply(
                                instance_location.wrapped.Transformation()
                            )
                            port_location = Location(trsf)
                            # pc_logging.debug(
                            #     "Result location: %s" % port_location
                            # )
                        # pc_logging.debug(
                        #     "Inherited port from %s to %s at %s: %s"
                        #     % (
                        #         interface_name,
                        #         instance_name,
                        #         self.name,
                        #         inherited_port_name,
                        #     )
                        # )
                        self.ports[inherited_port_name] = InterfacePort(
                            inherited_port_name,
                            self.project,
                            sketch=port.sketch,
                            location=port_location,
                        )

                    # TODO(clairbee): prepend the instance name to the param name
                    # TODO(clairbee): prepend only if it's not the only instance?
                    # pc_logging.debug(
                    #     "Inherited parameters: %s"
                    #     % str(inherit.interface.params)
                    # )
                    for (
                        param_name,
                        param,
                    ) in inherit.interface.params.items():
                        self.params[param_name] = param
                    # pc_logging.debug("Result parameters: %s" % str(self.params))

        # Enrich mating information
        mates = self.config.get("mates", None)
        if not mates is None:
            if self.abstract:
                pc_logging.error(
                    "Abstract interfaces cannot have mates: %s" % self.name
                )
                return

            if isinstance(mates, str):
                mates = {mates: {}}
            elif isinstance(mates, list):
                mates = {x: {} for x in mates}
            elif not isinstance(mates, dict):
                raise Exception(
                    "Invalid 'mates' section in the interface '%s'" % self.name
                )

            self.add_mates(self.project, mates)

    def add_mates(self, project, mates: dict):
        """Handles the "mates" sub-section of this interface's config,
        or the references to this interface in top level "mates" config sections
        of any project."""
        for target_interface_name, mate_target_config in mates.items():
            if not ":" in target_interface_name:
                target_interface_name = (
                    project.name + ":" + target_interface_name
                )
            target_package_name, short_target_interface_name = (
                resolve_resource_path(project.name, target_interface_name)
            )

            if target_package_name == project.name:
                target_project = project
            else:
                target_project = project.ctx.get_project(target_package_name)
            if target_project is None:
                pc_logging.error(
                    "Failed to find the target package for %s: %s"
                    % (target_interface_name, target_package_name)
                )
                continue
            target_interface = target_project.get_interface(
                short_target_interface_name
            )
            if target_interface is None:
                pc_logging.error(
                    "Failed to find the target interface: %s"
                    % target_interface_name
                )
                continue
            if target_interface.abstract:
                pc_logging.error(
                    "Cannot mate with an abstract interface: %s"
                    % target_interface_name
                )
                continue
            project.ctx.add_mate(self, target_interface, mate_target_config)

    def test(self):
        _ = self.get_ports()

    def info(self):
        info = {
            "name": self.name,
            "desc": self.desc,
            "ports": list(self.get_ports().values()),
            "parameters": list(self.params.values()),
            "inherits": self.get_parents(),
        }
        if self.abstract:
            info["abstract"] = True
        if self.lead_port is not None:
            info["leadPort"] = self.lead_port
        return info

    async def get_components(self):
        components = []
        for port in self.get_ports().values():
            components.append(port.location)
            if port.sketch is not None:
                sketch_components = list(await port.sketch.get_components())

                def move_component(component, move_components):
                    nonlocal port
                    if isinstance(component, list):
                        component = move_components(component)
                    else:
                        component = component.Located(port.location.wrapped)
                    return component

                def move_components(components):
                    components = list(
                        map(
                            lambda x: move_component(x, move_components),
                            components,
                        )
                    )
                    return components

                sketch_components = move_components(sketch_components)
                components.append(sketch_components)
        return components

    async def show_async(self):
        components = []
        try:
            components = await self.get_components()
        except Exception as e:
            pc_logging.error(e)

        if len(components) != 0:
            import importlib

            ocp_vscode = importlib.import_module("ocp_vscode")
            if ocp_vscode is None:
                pc_logging.warning(
                    'Failed to load "ocp_vscode". Giving up on connection to VS Code.'
                )
            else:
                try:
                    pc_logging.info('Visualizing in "OCP CAD Viewer"...')
                    ocp_vscode.show(
                        *components,
                        progress=None,
                        # TODO(clairbee): make showing (and the connection
                        # to ocp_vscode) a part of the context, and memorize
                        # which part was displayed last. Keep the camera
                        # if the part has not changed.
                        # reset_camera=ocp_vscode.Camera.KEEP,
                    )
                except Exception as e:
                    pc_logging.warning(e)
                    pc_logging.warning(
                        'No VS Code or "OCP CAD Viewer" extension detected.'
                    )

    def show(self):
        asyncio.run(self.show_async())



#
# OpenVMP, 2023
#
# Author: Roman Kuzmenko
# Created: 2023-08-19
#
# Licensed under Apache License, Version 2.0.
#

import typing

from .part import Part
from .shape_factory import ShapeFactory


class PartFactory(ShapeFactory):
    # TODO(clairbee): Make the next line work for part_factory_file only
    path: typing.Optional[str] = None
    part: Part
    name: str
    orig_name: str

    def __init__(
        self,
        ctx,
        source_project,
        target_project,
        config: object,
    ):
        super().__init__(ctx, source_project, config)
        self.target_project = target_project
        self.name = config["name"]
        self.orig_name = config["orig_name"]

    def _create_part(self, config: object) -> Part:
        part = Part(config)
        part.project_name = (
            self.target_project.name
        )  # TODO(clairbee): pass it via the constructor
        # TODO(clairbee): Make the next line work for part_factory_file only
        part.instantiate = lambda part_self: self.instantiate(part_self)
        part.info = lambda: self.info(part)
        part.with_ports = self.with_ports
        return part

    def _create(self, config: object):
        self.part = self._create_part(config)
        if self.path:
            self.part.path = self.path

        self.target_project.parts[self.name] = self.part
        self.ctx.stats_parts += 1



#
# OpenVMP, 2024
#
# Author: Roman Kuzmenko
# Created: 2024-04-20
#
# Licensed under Apache License, Version 2.0.
#

from OCP.gp import (
    gp_Circ,
    gp_Ax2,
    gp_Pnt,
    gp_Dir,
)
from OCP.BRepBuilderAPI import (
    BRepBuilderAPI_MakeEdge,
    BRepBuilderAPI_MakeWire,
    BRepBuilderAPI_MakeFace,
    BRepBuilderAPI_MakePolygon,
)
from OCP.ShapeFix import (
    ShapeFix_Face,
)

from .sketch_factory import SketchFactory
from . import logging as pc_logging


class Circle:
    x: float = 0.0
    y: float = 0.0
    radius: float = 0.0

    def __init__(self, config: object = {}):
        if isinstance(config, float) or isinstance(config, int):
            self.radius = config
        if isinstance(config, dict):
            self.x = config.get("x", 0.0)
            self.y = config.get("y", 0.0)
            self.radius = config.get("radius", 0.0)

    def to_wire(self):
        c = gp_Circ(
            gp_Ax2(
                gp_Pnt(self.x, self.y, 0),
                gp_Dir(0.0, 0.0, 1.0),
            ),
            self.radius,
        )
        edge = BRepBuilderAPI_MakeEdge(c)
        if edge.IsDone():
            return BRepBuilderAPI_MakeWire(edge.Edge()).Wire()
        return None


class Square:
    x: float = 0.0
    y: float = 0.0
    side: float = 0.0

    def __init__(self, config: object = {}):
        if isinstance(config, float) or isinstance(config, int):
            self.side = config
        if isinstance(config, dict):
            self.x = config.get("x", 0.0)
            self.y = config.get("y", 0.0)
            self.side = config.get("side", 0.0)

    def to_wire(self):
        polygon = BRepBuilderAPI_MakePolygon(
            gp_Pnt(
                self.x + self.side / 2.0,
                self.y + self.side / 2.0,
                0.0,
            ),
            gp_Pnt(
                self.x + self.side / 2.0,
                self.y - self.side / 2.0,
                0.0,
            ),
            gp_Pnt(
                self.x - self.side / 2.0,
                self.y - self.side / 2.0,
                0.0,
            ),
            gp_Pnt(
                self.x - self.side / 2.0,
                self.y + self.side / 2.0,
                0.0,
            ),
            True,
        )
        if polygon.IsDone():
            return polygon.Wire()
        return None


class Rect:
    x: float = 0.0
    y: float = 0.0
    side_x: float = 0.0
    side_y: float = 0.0

    def __init__(self, config: object = {}):
        if isinstance(config, dict):
            self.x = config.get("x", 0.0)
            self.y = config.get("y", 0.0)
            self.side_x = config.get("side-x", 0.0)
            self.side_y = config.get("side-y", 0.0)

    def to_wire(self):
        polygon = BRepBuilderAPI_MakePolygon(
            gp_Pnt(
                self.x + self.side_x / 2.0,
                self.y + self.side_y / 2.0,
                0.0,
            ),
            gp_Pnt(
                self.x + self.side_x / 2.0,
                self.y - self.side_y / 2.0,
                0.0,
            ),
            gp_Pnt(
                self.x - self.side_x / 2.0,
                self.y - self.side_y / 2.0,
                0.0,
            ),
            gp_Pnt(
                self.x - self.side_x / 2.0,
                self.y + self.side_y / 2.0,
                0.0,
            ),
            True,
        )
        if polygon.IsDone():
            return polygon.Wire()
        return None


# TODO(clairbee): distinguish between inner and outer wires?
#                 see Face::_make_from_wires in build123d/topology.py
class SketchFactoryBasic(SketchFactory):
    outer_circle: Circle = None
    outer_square: Square = None
    outer_rect: Rect = None
    inner_circles: list[Circle] = []
    inner_squares: list[Square] = []
    inner_rects: list[Rect] = []

    def __init__(self, ctx, source_project, target_project, config):
        with pc_logging.Action(
            "InitBasic", target_project.name, config["name"]
        ):
            super().__init__(
                ctx,
                source_project,
                target_project,
                config,
            )

            if "circle" in config:
                self.outer_circle = Circle(config["circle"])
            if "square" in config:
                self.outer_square = Square(config["square"])
            if "rectangle" in config:
                self.outer_rect = Rect(config["rectangle"])

            if "inner" in config:
                inner = config["inner"]
                if "circle" in inner:
                    self.inner_circles = [Circle(inner["circle"])]
                if "circles" in inner:
                    self.inner_circles.extend(
                        [Circle(c) for c in inner["circles"]]
                    )
                if "square" in inner:
                    self.inner_squares = [Square(inner["square"])]
                if "squares" in inner:
                    self.inner_squares.extend(
                        [Square(s) for s in inner["squares"]]
                    )
                if "rectangle" in inner:
                    self.inner_rects = [Rect(inner["rectangle"])]
                if "rectangles" in inner:
                    self.inner_rects.extend(
                        [Rect(r) for r in inner["rectangles"]]
                    )

            self._create(config)

    async def instantiate(self, sketch):
        with pc_logging.Action("Basic", sketch.project_name, sketch.name):
            shape = None

            try:
                # Get the outer wire
                outer_wire = None
                if not self.outer_circle is None:
                    outer_wire = self.outer_circle.to_wire()
                if not self.outer_square is None:
                    outer_wire = self.outer_square.to_wire()
                if not self.outer_rect is None:
                    outer_wire = self.outer_rect.to_wire()

                # # Fix the outer wire
                # wire_fixer = ShapeFix_Shape(outer_wire)
                # wire_fixer.Perform()
                # fixed_outer_wire_shape = wire_fixer.Shape()
                # outer_wire = TopoDS.Wire_s(fixed_outer_wire_shape)

                # Collect inner wires
                inner_wires = []
                for inner_circle in self.inner_circles:
                    inner_wire = inner_circle.to_wire()
                    if not inner_wire is None:
                        inner_wires.append(inner_wire)
                for inner_square in self.inner_squares:
                    inner_wire = inner_square.to_wire()
                    if not inner_wire is None:
                        inner_wires.append(inner_wire)
                for inner_rect in self.inner_rects:
                    inner_wire = inner_rect.to_wire()
                    if not inner_wire is None:
                        inner_wires.append(inner_wire)

                # Build the face
                face_builder = BRepBuilderAPI_MakeFace(outer_wire, True)
                for inner_wire in inner_wires:
                    if not inner_wire.is_closed:
                        raise ValueError(
                            "Cannot build face(s): inner wire is not closed"
                        )
                    face_builder.Add(inner_wire.wrapped)

                face_builder.Build()
                if not face_builder.IsDone():
                    raise ValueError(
                        f"Cannot build face(s): {face_builder.Error()}"
                    )

                face = face_builder.Face()

                # # Fix the face
                face_fixer = ShapeFix_Face(face)
                face_fixer.FixOrientation()
                face_fixer.Perform()
                shape = face_fixer.Result()
            except Exception as e:
                pc_logging.exception("Failed to create a basic sketch: %s" % e)

            self.ctx.stats_sketches_instantiated += 1

            pc_logging.debug("Basic sketch instantiated")
            return shape



#
# OpenVMP, 2024
#
# Author: Roman Kuzmenko
# Created: 2024-09-07
#
# Licensed under Apache License, Version 2.0.
#

import os

from .provider_factory import ProviderFactory
from .file_factory import FileFactory
from . import factory
from . import logging as pc_logging


class ProviderFactoryFile(ProviderFactory):
    path: str = None
    extension: str
    fileFactory: FileFactory = None

    def __init__(
        self,
        ctx,
        source_project,
        target_project,
        config,
        extension="",
        can_create=False,
    ):
        super().__init__(ctx, source_project, target_project, config)
        self.extension = extension

        if "fileFrom" in config:
            self.fileFactory = factory.instantiate(
                "file", config["fileFrom"], ctx, self.project, config
            )

        if "path" in config:
            self.path = config["path"]
        else:
            self.path = self.orig_name + extension

        if not os.path.isdir(source_project.config_dir):
            raise Exception(
                "ERROR: The project config directory must be a directory, found: '%s'"
                % source_project.config_dir
            )
        self.path = os.path.join(source_project.config_dir, self.path)

        # self.filePath is what file to download if necessary
        if "filePath" in config:
            self.filePath = os.path.join(
                source_project.config_dir, config["filePath"]
            )
        else:
            self.filePath = self.path

        if self.fileFactory is None or self.filePath != self.path:
            # If the user did not supply a way to download the script,
            # check if the script exists
            exists = os.path.exists(self.path)
            if not can_create and not exists:
                raise Exception(
                    "ERROR: The provider path (%s) must exist" % self.path
                )
            if exists and not os.path.isfile(self.path):
                raise Exception(
                    "ERROR: The provider path (%s) must be a file" % self.path
                )

    async def prepare_script(self, provider) -> bool:
        if not self.fileFactory is None and not os.path.exists(self.filePath):
            with pc_logging.Action(
                "File", self.target_project.name, provider.name
            ):
                await self.fileFactory.download(self.filePath)

        # Check if the data file is present
        if (
            not os.path.exists(self.filePath)
            or os.path.getsize(self.filePath) == 0
        ):
            pc_logging.error(
                "Provider data file is empty or does not exist: %s"
                % self.filePath
            )
            return False

        # Check if the script file is present
        if not os.path.exists(self.path) or os.path.getsize(self.path) == 0:
            pc_logging.error(
                "Provider script is empty or does not exist: %s" % self.path
            )
            return False

        return True



#
# OpenVMP, 2023
#
# Author: Roman Kuzmenko
# Created: 2023-12-30
#
# Licensed under Apache License, Version 2.0.

import asyncio
import hashlib
import os
import pathlib
import subprocess
import sys
import threading

from . import runtime
from . import logging as pc_logging


class PythonRuntime(runtime.Runtime):
    def __init__(self, ctx, sandbox, version=None):
        if version is None:
            version = "%d.%d" % (sys.version_info.major, sys.version_info.minor)
        super().__init__(ctx, "python-" + sandbox + "-" + version)
        self.version = version

        # Runtimes are meant to be executed from dedicated threads, outside of
        # the asyncio event loop. So a threading lock is appropriate here.
        self.lock = threading.RLock()
        self.tls = threading.local()

    def get_async_lock(self):
        if not hasattr(self.tls, "async_locks"):
            self.tls.async_locks = {}
        self_id = id(self)
        if self_id not in self.tls.async_locks:
            self.tls.async_locks[self_id] = asyncio.Lock()
        return self.tls.async_locks[self_id]

    def once(self):
        pass

    async def run(self, cmd, stdin="", cwd=None):
        pc_logging.debug("Running: %s", cmd)
        p = await asyncio.create_subprocess_exec(
            # cmd,
            cmd[0],
            *cmd[1:],
            stdin=subprocess.PIPE,
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
            shell=False,
            # TODO(clairbee): creationflags=subprocess.CREATE_NO_WINDOW,
            cwd=cwd,
        )
        stdout, stderr = await p.communicate(
            input=stdin.encode(),
            # TODO(clairbee): add timeout
        )

        stdout = stdout.decode()
        stderr = stderr.decode()

        # if stdout:
        #     pc_logging.debug("Output of %s: %s" % (cmd, stdout))
        # if stderr:
        #     pc_logging.debug("Error of %s: %s" % (cmd, stderr))

        # TODO(clairbee): remove the below when a better troubleshooting mechanism is introduced
        # f = open("/tmp/log", "w")
        # f.write("Completed: %s\n" % cmd)
        # f.write(" stdin: %s\n" % stdin)
        # f.write(" stderr: %s\n" % stderr)
        # f.write(" stdout: %s\n" % stdout)
        # f.close()

        return stdout, stderr

    async def ensure(self, python_package):
        self.once()

        # TODO(clairbee): expire the guard file after a certain time

        python_package_hash = hashlib.sha256(
            python_package.encode()
        ).hexdigest()
        guard_path = os.path.join(
            self.path, ".partcad.installed." + python_package_hash
        )
        with self.lock:
            async with self.get_async_lock():
                if not os.path.exists(guard_path):
                    with pc_logging.Action(
                        "PipInst", self.version, python_package
                    ):
                        await self.run(["-m", "pip", "install", python_package])
                    pathlib.Path(guard_path).touch()

    async def prepare_for_package(self, project):
        self.once()

        # TODO(clairbee): expire the guard file after a certain time

        # Check if this project has python requirements
        requirements_path = os.path.join(project.path, "requirements.txt")
        if os.path.exists(requirements_path):
            # See if it was already prepared once
            project_hash = hashlib.sha256(project.path.encode()).hexdigest()
            flag_filename = ".partcad.project." + project_hash
            flag_path = os.path.join(self.path, flag_filename)
            with self.lock:
                async with self.get_async_lock():
                    if not os.path.exists(flag_path) or os.path.getmtime(
                        requirements_path
                    ) > os.path.getmtime(flag_path):
                        # Install requirements and remember when we did that
                        with pc_logging.Action(
                            "PipReqs", self.version, project.name
                        ):
                            await self.run(
                                [
                                    "-m",
                                    "pip",
                                    "install",
                                    "-r",
                                    requirements_path,
                                ]
                            )
                        pathlib.Path(flag_path).touch()

        # Install dependencies of the package
        if "pythonRequirements" in project.config_obj:
            for req in project.config_obj["pythonRequirements"]:
                await self.ensure(req)

    async def prepare_for_shape(self, config):
        self.once()

        # Install dependencies of this part
        if "pythonRequirements" in config:
            for req in config["pythonRequirements"]:
                await self.ensure(req)



#
# OpenVMP, 2023
#
# Author: Roman Kuzmenko
# Created: 2023-12-30
#
# Licensed under Apache License, Version 2.0.

import os

from .user_config import user_config


class Runtime:
    @staticmethod
    def get_internal_state_dir():
        return os.path.join(
            user_config.internal_state_dir,
            "runtime",
        )

    def __init__(self, ctx, name):
        self.ctx = ctx
        self.name = name
        self.path = os.path.join(
            Runtime.get_internal_state_dir(),
            "partcad-" + name,  # Leave "partcad" for UX (e.g. in VS Code)
        )
        self.initialized = os.path.exists(self.path)



#
# OpenVMP, 2024
#
# Author: Roman Kuzmenko
# Created: 2024-08-24
#
# Licensed under Apache License, Version 2.0.
#

import base64
import importlib
import httpx
from path import Path
import re
import threading
import time
from typing import Any

# Lazy-load AI imports as they are not always needed
# import ollama
ollama = None

from . import logging as pc_logging
from .user_config import user_config

lock = threading.Lock()

model_tokens = {}

ollama_num_thread = None

models_pulled = {}


def ollama_once():
    global ollama, ollama_num_thread

    with lock:
        if ollama is None:
            ollama = importlib.import_module("ollama")

            ollama_num_thread = user_config.ollama_num_thread

    return True


def model_once(model: str):
    global models_pulled

    ollama_once()

    with lock:
        if not model in models_pulled:
            with pc_logging.Action("OllamaPull", model):
                ollama.pull(model)
                models_pulled[model] = True


class AiOllama:
    def generate_ollama(
        self,
        model: str,
        prompt: str,
        config: dict[str, Any] = {},
        options_num: int = 1,
    ):
        model_once(model)

        pc_logging.info(
            "Generating with Ollama: asking for %d alternatives", options_num
        )

        if not ollama_once():
            return None

        image_content = []

        def insert_image(match):
            filename = match.group(1)
            image_index = len(image_content)
            image_content.append(
                Path(filename).read_bytes(),
                # base64.b64encode(Path(filename).read_bytes()).decode(),
            )
            return f"The attached image number {image_index}.\n"

        prompt = re.sub(r"INSERT_IMAGE_HERE\(([^)]*)\)", insert_image, prompt)

        if "tokens" in config:
            tokens = config["tokens"]
        else:
            tokens = model_tokens[model] if model in model_tokens else None

        if "top_p" in config:
            top_p = config["top_p"]
        else:
            top_p = None

        if "top_k" in config:
            top_k = config["top_k"]
        else:
            top_k = None

        if "temperature" in config:
            temperature = config["temperature"]
        else:
            temperature = None

        candidates = []
        for _ in range(options_num):
            retry = True
            while retry == True:
                retry = False
                try:
                    options = ollama.Options(
                        tokens=tokens,
                        num_thread=ollama_num_thread,
                        top_p=top_p,
                        top_k=top_k,
                        temperature=temperature,
                    )
                    pc_logging.debug("Prompt: %s" % prompt)
                    pc_logging.debug("Images: %d" % len(image_content))
                    response = ollama.generate(
                        model=model,
                        context=[],  # do not accumulate context uncontrollably
                        prompt=prompt,
                        images=image_content,
                        options=options,
                    )
                except httpx.ConnectError as e:
                    pc_logging.exception(e)
                    pc_logging.error(
                        "Failed to connect to Ollama. Is it running?"
                    )
                    retry = True
                    time.sleep(15)
                except ollama._types.ResponseError as e:
                    pc_logging.exception(e)
                    pc_logging.error(
                        "Failed to generate with Ollama: %s" % str(e)
                    )
                    pc_logging.warning(
                        f"Consider running 'ollama run {model}' first..."
                    )
                    if "not found" in str(e):
                        retry = True
                        time.sleep(15)
                    else:
                        time.sleep(1)
                except Exception as e:
                    pc_logging.exception(e)
                    retry = True
                    time.sleep(1)

            if not response or not "response" in response:
                error = "%s: Failed to generate content" % self.name
                pc_logging.error(error)
                return

            pc_logging.info("Response: %s", str(response))
            # Perform Ollama-specific output sanitization
            response_text = response["response"]
            response_text = response_text.replace("\\'", "'")
            candidates.append(response_text)

        products = []
        try:
            for candidate in candidates:
                if candidate:
                    products.append(candidate)
        except Exception as e:
            pc_logging.exception(e)

        return products



#
# OpenVMP, 2024
#
# Author: Roman Kuzmenko
# Created: 2024-02-09
#
# Licensed under Apache License, Version 2.0.

import asyncio
from concurrent.futures import ThreadPoolExecutor
import os
import sys

cpu_count = os.cpu_count()
# Leave one core for the asyncio event loop and stuff
if cpu_count > 1:
    cpu_count -= 1
executor = ThreadPoolExecutor(cpu_count, "partcad-executor-")


# run returns a future to wait for 'method' to be completed in a separate thread
async def run(method, *args):
    global executor

    if (
        sys.version_info[0] == 3 and sys.version_info[1] >= 10
    ) or sys.version_info[0] > 3:
        return await asyncio.get_running_loop().run_in_executor(
            executor, method, *args
        )
    else:
        # Python 3.9 and lower has a buggy '.run_in_executor()'.
        # It crashes with:
        # `got Future <Future pending> attached to a different loop`
        # TODO(clairbee): consider simply spawning a thread and waiting for completion
        return method(*args)


# run returns a future to wait for 'coroutine' to be completed in a separate thread
async def run_async(coroutine, *args):
    global executor

    def method():
        return asyncio.run(coroutine(*args))

    if (
        sys.version_info[0] == 3 and sys.version_info[1] >= 10
    ) or sys.version_info[0] > 3:
        return await asyncio.get_running_loop().run_in_executor(
            executor, method
        )
    else:
        # Python 3.9 and lower has a buggy '.run_in_executor()'.
        # It crashes with:
        # `got Future <Future pending> attached to a different loop`
        # TODO(clairbee): consider simply spawning a thread and waiting for completion
        return asyncio.run(coroutine(*args))



#
# OpenVMP, 2024
#
# Author: Roman Kuzmenko
# Created: 2024-01-06
#
# Licensed under Apache License, Version 2.0.
#

import build123d as b3d

from .part_factory_file import PartFactoryFile
from . import logging as pc_logging


class PartFactory3mf(PartFactoryFile):
    def __init__(self, ctx, source_project, target_project, config):
        with pc_logging.Action("Init3MF", target_project.name, config["name"]):
            super().__init__(
                ctx,
                source_project,
                target_project,
                config,
                extension=".3mf",
            )
            # Complement the config object here if necessary
            self._create(config)

    async def instantiate(self, part):
        await super().instantiate(part)

        with pc_logging.Action("3MF", part.project_name, part.name):
            shape = b3d.Mesher().read(self.path)[0].wrapped

            self.ctx.stats_parts_instantiated += 1

            return shape



#
# OpenVMP, 2023
#
# Author: Roman Kuzmenko
# Created: 2024-01-22
#
# Licensed under Apache License, Version 2.0.
#

import os

from .part_factory import PartFactory
from . import logging as pc_logging


class PartFactoryFile(PartFactory):
    extension: str

    def __init__(
        self,
        ctx,
        source_project,
        target_project,
        config,
        extension="",
        can_create=False,
    ):
        super().__init__(ctx, source_project, target_project, config)
        self.extension = extension

        if "path" in config:
            self.path = config["path"]
        else:
            self.path = self.orig_name + extension

        if not os.path.isdir(source_project.config_dir):
            raise Exception(
                "ERROR: The project config directory must be a directory, found: '%s'"
                % source_project.config_dir
            )
        self.path = os.path.join(source_project.config_dir, self.path)

        if self.fileFactory is None:
            # If the user did not supply a way to download the file,
            # check if the file exists
            exists = os.path.exists(self.path)
            if not can_create and not exists:
                raise Exception(
                    "ERROR: The part path (%s) must exist" % self.path
                )
            if exists and not os.path.isfile(self.path):
                raise Exception(
                    "ERROR: The part path (%s) must be a file" % self.path
                )

    async def instantiate(self, part):
        if not self.fileFactory is None and not os.path.exists(part.path):
            with pc_logging.Action("File", self.target_project.name, part.name):
                await self.fileFactory.download(part.path)



#
# OpenVMP, 2023
#
# Author: Roman Kuzmenko
# Created: 2024-02-19
#
# Licensed under Apache License, Version 2.0.
#

from . import factory

from .file_factory import FileFactory
from .port import WithPorts


class ShapeFactory(factory.Factory):
    fileFactory: FileFactory

    def __init__(self, ctx, project, config) -> None:
        super().__init__()

        self.ctx = ctx
        self.project = project
        self.config = config

        if "fileFrom" in config:
            self.fileFactory = factory.instantiate(
                "file", config["fileFrom"], ctx, project, project, config
            )
        else:
            self.fileFactory = None

        self.with_ports = WithPorts(config["name"], project, config)

    def info(self, shape):
        """This is the default implementation of the get_info method for factories."""
        return shape.shape_info()



#
# OpenVMP, 2023-2024
#
# Author: Roman Kuzmenko
# Created: 2024-09-07
#
# Licensed under Apache License, Version 2.0.
#

# This script is executed within the python sandbox environment (python runtime)
# to invoke `build1213d` scripts.

import os
import runpy
import sys

sys.path.append(os.path.dirname(__file__))
import wrapper_common


def process(path, request):
    try:
        result = runpy.run_path(
            path,
            init_globals={
                "request": request,
            },
            run_name=request["api"],
        )
    except Exception as e:
        result = {"exception": e}

    if "output" in result:
        output = result["output"]
    else:
        output = {}

    if "exception" in result:
        output["exception"] = result["exception"]
        wrapper_common.handle_exception(result["exception"], path)

    return output


path, request = wrapper_common.handle_input()

# Call the API endpoint
result = process(path, request)

wrapper_common.handle_output(result)



#
# OpenVMP, 2024
#
# Author: Roman Kuzmenko
# Created: 2024-02-04
#
# Licensed under Apache License, Version 2.0.

import logging
from logging import DEBUG, INFO, WARN, WARNING, ERROR, CRITICAL
import threading
import time

# Track if any errors occurred during the execution for test purposes and for
# the main program to know if it should exit with an error code.
had_errors = False

# Other packages (such as 'pip' or 'pytest') may mess with the root logger
# in a way that we don't want. We create a separate logger for our own use.
# For example, 'pip' sets up a filter that drops 'extra's from log records,
# thus turning 'process_start' and 'process_end' into regular error messages.
logging.getLogger("partcad").propagate = False

# Wrap the logger to make it easier to use
setLevel = lambda *a, **kw: logging.getLogger("partcad").setLevel(*a, **kw)
getLevel = lambda: logging.getLogger("partcad").level
log = lambda *a, **kw: logging.getLogger("partcad").log(*a, **kw)
debug = lambda *a, **kw: logging.getLogger("partcad").debug(*a, **kw)
info = lambda *a, **kw: logging.getLogger("partcad").info(*a, **kw)
warn = lambda *a, **kw: logging.getLogger("partcad").warn(*a, **kw)
warning = lambda *a, **kw: logging.getLogger("partcad").warning(*a, **kw)


def error(*args, **kwargs):
    global had_errors
    had_errors = True
    logging.getLogger("partcad").error(*args, **kwargs)


def critical(*args, **kwargs):
    global had_errors
    had_errors = True
    logging.getLogger("partcad").critical(*args, **kwargs)


# Some pytest versions/configurations/plugins mess with the exception method
# so lambdas don't work
def exception(
    *args,
):
    global had_errors
    had_errors = True
    logging.getLogger("partcad").exception(*args)


# Create 'ops' that are used for dependency injection of the logic to control
# the logging context (e.g. the current state of processes and actions).
def default_process_start(
    self_ops, op: str, package: str, item: str | None = None
):
    if item is None:
        info("Starting process: %s: %s" % (op, package))
    else:
        info("Starting process: %s: %s: %s" % (op, package, item))


def default_process_end(self_ops, op: str, package: str, item: str = None):
    pass


def default_action_start(
    self_ops, op: str, package: str, item: str | None = None
):
    if item is None:
        info("Starting action: %s: %s" % (op, package))
    else:
        info("Starting action: %s: %s: %s" % (op, package, item))


def default_action_end(self_ops, op: str, package: str, item: str = None):
    pass


# Dependency injection point for logging plugins
class Ops:
    process_start = default_process_start
    process_end = default_process_end
    action_start = default_action_start
    action_end = default_action_end


ops = Ops()


# Only one 'process' at a time, no recursion
process_lock = threading.Lock()


# Classes to be used with "with()" to alter the logging context.
class Process(object):
    def __init__(self, op: str, package: str, item: str = None):
        self.op = op
        self.package = package
        self.item = item
        self.succeeded = False
        self.start = 0.0

    async def __aenter__(self):
        self.__enter__()

    def __enter__(self):
        global process_lock

        if process_lock.acquire():
            self.start = time.time()
            ops.process_start(self.op, self.package, self.item)
            self.succeeded = True
        else:
            error("Nested process is detected. Status reporting is invalid.")
            self.succeeded = False

    async def __aexit__(self, *args):
        self.__exit__(*args)

    def __exit__(self, *_args):
        global process_lock
        global info

        if self.succeeded:
            process_lock.release()
            ops.process_end(self.op, self.package, self.item)

            delta = time.time() - self.start
            if self.item is None:
                info("DONE: %s: %s: %.2fs" % (self.op, self.package, delta))
            else:
                info(
                    "DONE: %s: %s: %s: %.2fs"
                    % (self.op, self.package, self.item, delta)
                )


class Action(object):
    def __init__(
        self, op: str, package: str, item: str = None, extra: str = None
    ):
        self.op = op
        self.package = package
        if extra:
            self.item = item + " : " + extra
        else:
            self.item = item

    async def __aenter__(self):
        self.__enter__()

    def __enter__(self):
        ops.action_start(self.op, self.package, self.item)

    async def __aexit__(self, *args):
        self.__exit__(*args)

    def __exit__(self, *_args):
        ops.action_end(self.op, self.package, self.item)
